import pytest

import numpy.testing as npt
from xarray import Dataset, DataArray, concat, open_dataset, open_mfdataset
import xarray.testing as xrt

import dask.array
import numpy as np
from pathlib import Path
from scipy.integrate import quad_vec

from xbout.tests.test_load import bout_xyt_example_files, create_bout_ds
from xbout.tests.test_region import (
    params_guards,
    params_guards_values,
    params_boundaries,
    params_boundaries_values,
)
from xbout import BoutDatasetAccessor, open_boutdataset
from xbout.geometries import apply_geometry
from xbout.utils import _set_attrs_on_all_vars, _1d_coord_from_spacing
from xbout.tests.utils_for_tests import set_geometry_from_input_file


EXAMPLE_OPTIONS_FILE_PATH = "./xbout/tests/data/options/BOUT.inp"


class TestBoutDatasetIsXarrayDataset:
    """
    Set of tests to check that BoutDatasets behave similarly to xarray Datasets.
    (With the accessor approach these should pass trivially now.)
    """

    def test_concat(self, bout_xyt_example_files):
        dataset_list1 = bout_xyt_example_files(None, nxpe=3, nype=4, nt=1)
        with pytest.warns(UserWarning):
            bd1 = open_boutdataset(
                datapath=dataset_list1, inputfilepath=None, keep_xboundaries=False
            )
        dataset_list2 = bout_xyt_example_files(None, nxpe=3, nype=4, nt=1)
        with pytest.warns(UserWarning):
            bd2 = open_boutdataset(
                datapath=dataset_list2, inputfilepath=None, keep_xboundaries=False
            )
        result = concat([bd1, bd2], dim="run")
        assert result.dims == {**bd1.dims, "run": 2}

    def test_isel(self, bout_xyt_example_files):
        dataset_list = bout_xyt_example_files(None, nxpe=1, nype=1, nt=1)
        with pytest.warns(UserWarning):
            bd = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )
        actual = bd.isel(x=slice(None, None, 2))
        expected = bd.bout.data.isel(x=slice(None, None, 2))
        xrt.assert_equal(actual, expected)


class TestBoutDatasetMethods:
    def test_get_field_aligned(self, bout_xyt_example_files):
        dataset_list = bout_xyt_example_files(None, nxpe=3, nype=4, nt=1)
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["x"]
        ds["Rxy"] = ds["x"]
        ds["Zxy"] = ds["y"]

        ds = apply_geometry(ds, "toroidal")

        n = ds["n"]
        n.attrs["direction_y"] = "Standard"
        n_aligned_from_array = n.bout.to_field_aligned()

        # check n_aligned does not exist yet
        assert "n_aligned" not in ds

        n_aligned_from_ds = ds.bout.get_field_aligned("n")
        xrt.assert_allclose(n_aligned_from_ds, n_aligned_from_array)
        xrt.assert_allclose(ds["n_aligned"], n_aligned_from_array)

        # check getting the cached version
        ds["n_aligned"] = ds["T"]
        ds["n_aligned"].attrs["direction_y"] = "Aligned"
        xrt.assert_allclose(ds.bout.get_field_aligned("n"), ds["T"])

    @pytest.mark.parametrize("mxg", [0, pytest.param(2, marks=pytest.mark.long)])
    @pytest.mark.parametrize("myg", [pytest.param(0, marks=pytest.mark.long), 2])
    def test_remove_yboundaries(self, bout_xyt_example_files, mxg, myg):
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=1,
            nype=6,
            nt=1,
            grid="grid",
            guards={"x": mxg, "y": myg},
            topology="connected-double-null",
            syn_data_type="linear",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_yboundaries=True,
        )

        dataset_list_no_yboundaries, grid_ds_no_yboundaries = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=1,
            nype=6,
            nt=1,
            grid="grid",
            guards={"x": mxg, "y": 0},
            topology="connected-double-null",
            syn_data_type="linear",
        )

        ds_no_yboundaries = open_boutdataset(
            datapath=dataset_list_no_yboundaries,
            gridfilepath=grid_ds_no_yboundaries,
            geometry="toroidal",
            keep_yboundaries=False,
        )

        ds = ds.bout.remove_yboundaries()

        assert ds.metadata["keep_yboundaries"] == 0
        for v in ds:
            assert ds[v].metadata["keep_yboundaries"] == 0

        # expect theta coordinate to be different, so ignore
        ds = ds.drop_vars("theta")
        ds_no_yboundaries = ds_no_yboundaries.drop_vars("theta")

        # expect MYG can be different, so ignore
        del ds.metadata["MYG"]
        del ds_no_yboundaries.metadata["MYG"]
        for v in ds:
            del ds[v].metadata["MYG"]
            # metadata on variables is a reference to ds_no_yboundaries.metadata, so
            # don't need to delete MYG again

        xrt.assert_equal(ds, ds_no_yboundaries)

    @pytest.mark.parametrize(
        "nz",
        [
            pytest.param(6, marks=pytest.mark.long),
            7,
            pytest.param(8, marks=pytest.mark.long),
            pytest.param(9, marks=pytest.mark.long),
        ],
    )
    def test_to_field_aligned(self, bout_xyt_example_files, nz):
        dataset_list = bout_xyt_example_files(
            None, lengths=(3, 3, 4, nz), nxpe=1, nype=1, nt=1
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["x"]
        ds["Rxy"] = ds["x"]
        ds["Zxy"] = ds["y"]

        ds = apply_geometry(ds, "toroidal")

        # set up test variable
        n = ds["n"].load()
        zShift = ds["zShift"].load()
        for t in range(ds.sizes["t"]):
            for x in range(ds.sizes["x"]):
                for y in range(ds.sizes["theta"]):
                    zShift[x, y] = (
                        (x * ds.sizes["theta"] + y) * 2.0 * np.pi / ds.sizes["zeta"]
                    )
                    for z in range(nz):
                        n[t, x, y, z] = 1000.0 * t + 100.0 * x + 10.0 * y + z

        n.attrs["direction_y"] = "Standard"
        ds["n"] = n

        assert ds["n"].direction_y == "Standard"

        # Create field-aligned Dataset
        ds_al = ds.bout.to_field_aligned()

        n_al = ds_al["n"]

        assert n_al.direction_y == "Aligned"

        for t in range(ds.sizes["t"]):
            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 0, z].values,
                    1000.0 * t + z % nz,
                    rtol=1.0e-15,
                    atol=5.0e-16,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 1, z].values,
                    1000.0 * t + 10.0 * 1.0 + (z + 1) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 2, z].values,
                    1000.0 * t + 10.0 * 2.0 + (z + 2) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 3, z].values,
                    1000.0 * t + 10.0 * 3.0 + (z + 3) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 0, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 0.0 + (z + 4) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 1, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 1.0 + (z + 5) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 2, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 2.0 + (z + 6) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 3, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 3.0 + (z + 7) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

    def test_to_field_aligned_dask(self, bout_xyt_example_files):

        nz = 6

        dataset_list = bout_xyt_example_files(
            None, lengths=(3, 3, 4, nz), nxpe=1, nype=1, nt=1
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["x"]
        ds["Rxy"] = ds["x"]
        ds["Zxy"] = ds["y"]

        ds = apply_geometry(ds, "toroidal")

        # set up test variable
        n = ds["n"].load()
        zShift = ds["zShift"].load()
        for t in range(ds.sizes["t"]):
            for x in range(ds.sizes["x"]):
                for y in range(ds.sizes["theta"]):
                    zShift[x, y] = (
                        (x * ds.sizes["theta"] + y) * 2.0 * np.pi / ds.sizes["zeta"]
                    )
                    for z in range(nz):
                        n[t, x, y, z] = 1000.0 * t + 100.0 * x + 10.0 * y + z

        # The above loop required the call to .load(), but that turned the data into a
        # numpy array. Now convert back to dask
        n = n.chunk({"t": 1})
        ds = ds.chunk({"t": 1})
        assert isinstance(n.data, dask.array.Array)
        assert isinstance(ds["n"].data, dask.array.Array)

        n.attrs["direction_y"] = "Standard"
        ds["n"] = n

        assert ds["n"].direction_y == "Standard"

        # Create field-aligned Dataset
        ds_al = ds.bout.to_field_aligned()

        n_al = ds_al["n"]

        assert n_al.direction_y == "Aligned"

        for t in range(ds.sizes["t"]):
            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 0, z].values,
                    1000.0 * t + z % nz,
                    rtol=1.0e-15,
                    atol=5.0e-16,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 1, z].values,
                    1000.0 * t + 10.0 * 1.0 + (z + 1) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 2, z].values,
                    1000.0 * t + 10.0 * 2.0 + (z + 2) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 0, 3, z].values,
                    1000.0 * t + 10.0 * 3.0 + (z + 3) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 0, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 0.0 + (z + 4) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 1, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 1.0 + (z + 5) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 2, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 2.0 + (z + 6) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_al[t, 1, 3, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 3.0 + (z + 7) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

    @pytest.mark.parametrize(
        "nz",
        [
            pytest.param(6, marks=pytest.mark.long),
            7,
            pytest.param(8, marks=pytest.mark.long),
            pytest.param(9, marks=pytest.mark.long),
        ],
    )
    def test_from_field_aligned(self, bout_xyt_example_files, nz):
        dataset_list = bout_xyt_example_files(
            None, lengths=(3, 3, 4, nz), nxpe=1, nype=1, nt=1
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["x"]
        ds["Rxy"] = ds["x"]
        ds["Zxy"] = ds["y"]

        ds = apply_geometry(ds, "toroidal")

        # set up test variable
        n = ds["n"].load()
        zShift = ds["zShift"].load()
        for t in range(ds.sizes["t"]):
            for x in range(ds.sizes["x"]):
                for y in range(ds.sizes["theta"]):
                    zShift[x, y] = (
                        (x * ds.sizes["theta"] + y) * 2.0 * np.pi / ds.sizes["zeta"]
                    )
                    for z in range(ds.sizes["zeta"]):
                        n[t, x, y, z] = 1000.0 * t + 100.0 * x + 10.0 * y + z

        n.attrs["direction_y"] = "Aligned"
        ds["n"] = n

        assert ds["n"].direction_y == "Aligned"

        # Create non-field-aligned Dataset
        ds_nal = ds.bout.from_field_aligned()

        n_nal = ds_nal["n"]

        assert n_nal.direction_y == "Standard"

        for t in range(ds.sizes["t"]):
            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 0, 0, z].values,
                    1000.0 * t + z % nz,
                    rtol=1.0e-15,
                    atol=5.0e-16,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 0, 1, z].values,
                    1000.0 * t + 10.0 * 1.0 + (z - 1) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 0, 2, z].values,
                    1000.0 * t + 10.0 * 2.0 + (z - 2) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 0, 3, z].values,
                    1000.0 * t + 10.0 * 3.0 + (z - 3) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 1, 0, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 0.0 + (z - 4) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 1, 1, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 1.0 + (z - 5) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 1, 2, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 2.0 + (z - 6) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

            for z in range(nz):
                npt.assert_allclose(
                    n_nal[t, 1, 3, z].values,
                    1000.0 * t + 100.0 * 1 + 10.0 * 3.0 + (z - 7) % nz,
                    rtol=1.0e-15,
                    atol=0.0,
                )  # noqa: E501

    @pytest.mark.parametrize("stag_location", ["CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"])
    def test_to_field_aligned_staggered(self, bout_xyt_example_files, stag_location):
        dataset_list = bout_xyt_example_files(
            None, lengths=(3, 3, 4, 8), nxpe=1, nype=1, nt=1
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["x"]
        ds["Rxy"] = ds["x"]
        ds["Zxy"] = ds["y"]

        ds = apply_geometry(ds, "toroidal")

        # set up test variable
        n = ds["n"].load()
        zShift = ds["zShift"].load()
        for t in range(ds.sizes["t"]):
            for x in range(ds.sizes["x"]):
                for y in range(ds.sizes["theta"]):
                    zShift[x, y] = (
                        (x * ds.sizes["theta"] + y) * 2.0 * np.pi / ds.sizes["zeta"]
                    )
                    for z in range(ds.sizes["zeta"]):
                        n[t, x, y, z] = 1000.0 * t + 100.0 * x + 10.0 * y + z

        ds["n"] = n

        assert ds["n"].direction_y == "Standard"

        # Create field-aligned Dataset
        ds_al = ds.bout.to_field_aligned()

        n_al = ds_al["n"].copy(deep=True)

        assert n_al.direction_y == "Aligned"

        # make 'n' staggered
        ds["n"].attrs["cell_location"] = stag_location

        if stag_location != "CELL_ZLOW":
            with pytest.raises(ValueError):
                # Check exception raised when needed zShift_CELL_*LOW is not present
                ds.bout.to_field_aligned()
            ds["zShift_" + stag_location] = zShift
            ds["zShift_" + stag_location].attrs["cell_location"] = stag_location
            ds = ds.set_coords("zShift_" + stag_location)

        # New field-aligned Dataset
        ds_al = ds.bout.to_field_aligned()

        n_stag_al = ds_al["n"]

        assert n_stag_al.direction_y == "Aligned"

        npt.assert_equal(n_stag_al.values, n_al.values)

    @pytest.mark.parametrize("stag_location", ["CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"])
    def test_from_field_aligned_staggered(self, bout_xyt_example_files, stag_location):
        dataset_list = bout_xyt_example_files(
            None, lengths=(3, 3, 4, 8), nxpe=1, nype=1, nt=1
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["x"]
        ds["Rxy"] = ds["x"]
        ds["Zxy"] = ds["y"]

        ds = apply_geometry(ds, "toroidal")

        # set up test variable
        n = ds["n"].load()
        zShift = ds["zShift"].load()
        for t in range(ds.sizes["t"]):
            for x in range(ds.sizes["x"]):
                for y in range(ds.sizes["theta"]):
                    zShift[x, y] = (
                        (x * ds.sizes["theta"] + y) * 2.0 * np.pi / ds.sizes["zeta"]
                    )
                    for z in range(ds.sizes["zeta"]):
                        n[t, x, y, z] = 1000.0 * t + 100.0 * x + 10.0 * y + z
        n.attrs["direction_y"] = "Aligned"
        ds["n"] = n
        ds["T"].attrs["direction_y"] = "Aligned"

        # Make non-field-aligned Dataset
        ds_nal = ds.bout.from_field_aligned()

        n_nal = ds_nal["n"].copy(deep=True)

        assert n_nal.direction_y == "Standard"

        # make 'n' staggered
        ds["n"].attrs["cell_location"] = stag_location

        if stag_location != "CELL_ZLOW":
            with pytest.raises(ValueError):
                # Check exception raised when needed zShift_CELL_*LOW is not present
                ds["n"].bout.from_field_aligned()
            ds["zShift_" + stag_location] = zShift
            ds["zShift_" + stag_location].attrs["cell_location"] = stag_location
            ds = ds.set_coords("zShift_" + stag_location)

        # New non-field-aligned Dataset
        ds_nal = ds.bout.from_field_aligned()

        n_stag_nal = ds_nal["n"]

        assert n_stag_nal.direction_y == "Standard"

        npt.assert_equal(n_stag_nal.values, n_nal.values)

    def test_set_parallel_interpolation_factor(self):
        ds = Dataset()
        ds["a"] = DataArray()
        ds = _set_attrs_on_all_vars(ds, "metadata", {})

        with pytest.raises(KeyError):
            ds.metadata["fine_interpolation_factor"]
        with pytest.raises(KeyError):
            ds["a"].metadata["fine_interpolation_factor"]

        ds.bout.fine_interpolation_factor = 42

        assert ds.metadata["fine_interpolation_factor"] == 42
        assert ds["a"].metadata["fine_interpolation_factor"] == 42

    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    @pytest.mark.parametrize(
        "vars_to_interpolate", [("n", "T"), pytest.param(..., marks=pytest.mark.long)]
    )
    def test_interpolate_parallel(
        self,
        bout_xyt_example_files,
        guards,
        keep_xboundaries,
        keep_yboundaries,
        vars_to_interpolate,
    ):
        # This test checks that the regions created in the new high-resolution Dataset by
        # interpolate_parallel are correct.
        # This test does not test the accuracy of the parallel interpolation (there are
        # other tests for that).

        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            guards=guards,
            grid="grid",
            topology="lower-disconnected-double-null",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        # Get high parallel resolution version of ds, and check that
        ds = ds.bout.interpolate_parallel(vars_to_interpolate)

        mxg = guards["x"]
        myg = guards["y"]

        if keep_xboundaries:
            ixs1 = ds.metadata["ixseps1"]
        else:
            ixs1 = ds.metadata["ixseps1"] - guards["x"]

        if keep_xboundaries:
            ixs2 = ds.metadata["ixseps2"]
        else:
            ixs2 = ds.metadata["ixseps2"] - guards["x"]

        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        jys11 = ds.metadata["jyseps1_1"] + ybndry
        jys21 = ds.metadata["jyseps2_1"] + ybndry
        ny_inner = ds.metadata["ny_inner"] + 2 * ybndry
        jys12 = ds.metadata["jyseps1_2"] + 3 * ybndry
        jys22 = ds.metadata["jyseps2_2"] + 3 * ybndry
        ny = ds.metadata["ny"] + 4 * ybndry

        for var in ["n", "T"]:
            v = ds[var]
            v_noregions = v.copy(deep=True)

            # Remove attributes that are expected to be different
            del v_noregions.attrs["regions"]

            v_lower_inner_PFR = v.bout.from_region("lower_inner_PFR")

            # Remove attributes that are expected to be different
            del v_lower_inner_PFR.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(x=slice(ixs1 + mxg), theta=slice(jys11 + 1)),
                v_lower_inner_PFR.isel(theta=slice(-myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys22 + 1, jys22 + 1 + myg)
                    ).values,
                    v_lower_inner_PFR.isel(theta=slice(-myg, None)).values,
                )

            v_lower_inner_intersep = v.bout.from_region("lower_inner_intersep")

            # Remove attributes that are expected to be different
            del v_lower_inner_intersep.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 - mxg, ixs2 + mxg), theta=slice(jys11 + 1)
                ),
                v_lower_inner_intersep.isel(theta=slice(-myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys11 + 1, jys11 + 1 + myg),
                    ).values,
                    v_lower_inner_intersep.isel(theta=slice(-myg, None)).values,
                )

            v_lower_inner_SOL = v.bout.from_region("lower_inner_SOL")

            # Remove attributes that are expected to be different
            del v_lower_inner_SOL.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(x=slice(ixs2 - mxg, None), theta=slice(jys11 + 1)),
                v_lower_inner_SOL.isel(theta=slice(-myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys11 + 1, jys11 + 1 + myg),
                    ).values,
                    v_lower_inner_SOL.isel(theta=slice(-myg, None)).values,
                )

            v_inner_core = v.bout.from_region("inner_core")

            # Remove attributes that are expected to be different
            del v_inner_core.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 + mxg), theta=slice(jys11 + 1, jys21 + 1)
                ),
                v_inner_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys22 + 1 - myg, jys22 + 1)
                    ).values,
                    v_inner_core.isel(theta=slice(myg)).values,
                )
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys12 + 1, jys12 + 1 + myg)
                    ).values,
                    v_inner_core.isel(theta=slice(-myg, None)).values,
                )

            v_inner_intersep = v.bout.from_region("inner_intersep")

            # Remove attributes that are expected to be different
            del v_inner_intersep.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 - mxg, ixs2 + mxg), theta=slice(jys11 + 1, jys21 + 1)
                ),
                v_inner_intersep.isel(theta=slice(myg, -myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys11 + 1 - myg, jys11 + 1),
                    ).values,
                    v_inner_intersep.isel(theta=slice(myg)).values,
                )
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys12 + 1, jys12 + 1 + myg),
                    ).values,
                    v_inner_intersep.isel(theta=slice(-myg, None)).values,
                )

            v_inner_sol = v.bout.from_region("inner_SOL")

            # Remove attributes that are expected to be different
            del v_inner_sol.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs2 - mxg, None), theta=slice(jys11 + 1, jys21 + 1)
                ),
                v_inner_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys11 + 1 - myg, jys11 + 1),
                    ).values,
                    v_inner_sol.isel(theta=slice(myg)).values,
                )
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys21 + 1, jys21 + 1 + myg),
                    ).values,
                    v_inner_sol.isel(theta=slice(-myg, None)).values,
                )

            v_upper_inner_PFR = v.bout.from_region("upper_inner_PFR")

            # Remove attributes that are expected to be different
            del v_upper_inner_PFR.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(x=slice(ixs1 + mxg), theta=slice(jys21 + 1, ny_inner)),
                v_upper_inner_PFR.isel(theta=slice(myg, None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys12 + 1 - myg, jys12 + 1)
                    ).values,
                    v_upper_inner_PFR.isel(theta=slice(myg)).values,
                )

            v_upper_inner_intersep = v.bout.from_region("upper_inner_intersep")

            # Remove attributes that are expected to be different
            del v_upper_inner_intersep.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 - mxg, ixs2 + mxg), theta=slice(jys21 + 1, ny_inner)
                ),
                v_upper_inner_intersep.isel(theta=slice(myg, None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys12 + 1 - myg, jys12 + 1),
                    ).values,
                    v_upper_inner_intersep.isel(theta=slice(myg)).values,
                )

            v_upper_inner_SOL = v.bout.from_region("upper_inner_SOL")

            # Remove attributes that are expected to be different
            del v_upper_inner_SOL.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs2 - mxg, None), theta=slice(jys21 + 1, ny_inner)
                ),
                v_upper_inner_SOL.isel(theta=slice(myg, None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys21 + 1 - myg, jys21 + 1),
                    ).values,
                    v_upper_inner_SOL.isel(theta=slice(myg)).values,
                )

            v_upper_outer_PFR = v.bout.from_region("upper_outer_PFR")

            # Remove attributes that are expected to be different
            del v_upper_outer_PFR.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(x=slice(ixs1 + mxg), theta=slice(ny_inner, jys12 + 1)),
                v_upper_outer_PFR.isel(theta=slice(-myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys21 + 1, jys21 + 1 + myg)
                    ).values,
                    v_upper_outer_PFR.isel(theta=slice(-myg, None)).values,
                )

            v_upper_outer_intersep = v.bout.from_region("upper_outer_intersep")

            # Remove attributes that are expected to be different
            del v_upper_outer_intersep.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 - mxg, ixs2 + mxg), theta=slice(ny_inner, jys12 + 1)
                ),
                v_upper_outer_intersep.isel(theta=slice(-myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys21 + 1, jys21 + 1 + myg),
                    ).values,
                    v_upper_outer_intersep.isel(theta=slice(-myg, None)).values,
                )

            v_upper_outer_SOL = v.bout.from_region("upper_outer_SOL")

            # Remove attributes that are expected to be different
            del v_upper_outer_SOL.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs2 - mxg, None), theta=slice(ny_inner, jys12 + 1)
                ),
                v_upper_outer_SOL.isel(theta=slice(-myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys12 + 1, jys12 + 1 + myg),
                    ).values,
                    v_upper_outer_SOL.isel(theta=slice(-myg, None)).values,
                )

            v_outer_core = v.bout.from_region("outer_core")

            # Remove attributes that are expected to be different
            del v_outer_core.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 + mxg), theta=slice(jys12 + 1, jys22 + 1)
                ),
                v_outer_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys21 + 1 - myg, jys21 + 1)
                    ).values,
                    v_outer_core.isel(theta=slice(myg)).values,
                )
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys11 + 1, jys11 + 1 + myg)
                    ).values,
                    v_outer_core.isel(theta=slice(-myg, None)).values,
                )

            v_outer_intersep = v.bout.from_region("outer_intersep")

            # Remove attributes that are expected to be different
            del v_outer_intersep.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 - mxg, ixs2 + mxg), theta=slice(jys12 + 1, jys22 + 1)
                ),
                v_outer_intersep.isel(theta=slice(myg, -myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys21 + 1 - myg, jys21 + 1),
                    ).values,
                    v_outer_intersep.isel(theta=slice(myg)).values,
                )
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys22 + 1, jys22 + 1 + myg),
                    ).values,
                    v_outer_intersep.isel(theta=slice(-myg, None)).values,
                )

            v_outer_sol = v.bout.from_region("outer_SOL")

            # Remove attributes that are expected to be different
            del v_outer_sol.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs2 - mxg, None), theta=slice(jys12 + 1, jys22 + 1)
                ),
                v_outer_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys12 + 1 - myg, jys12 + 1),
                    ).values,
                    v_outer_sol.isel(theta=slice(myg)).values,
                )
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys22 + 1, jys22 + 1 + myg),
                    ).values,
                    v_outer_sol.isel(theta=slice(-myg, None)).values,
                )

            v_lower_outer_PFR = v.bout.from_region("lower_outer_PFR")

            # Remove attributes that are expected to be different
            del v_lower_outer_PFR.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(x=slice(ixs1 + mxg), theta=slice(jys22 + 1, None)),
                v_lower_outer_PFR.isel(theta=slice(myg, None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 + mxg), theta=slice(jys11 + 1 - myg, jys11 + 1)
                    ).values,
                    v_lower_outer_PFR.isel(theta=slice(myg)).values,
                )

            v_lower_outer_intersep = v.bout.from_region("lower_outer_intersep")

            # Remove attributes that are expected to be different
            del v_lower_outer_intersep.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs1 - mxg, ixs2 + mxg), theta=slice(jys22 + 1, None)
                ),
                v_lower_outer_intersep.isel(theta=slice(myg, None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs1 - mxg, ixs2 + mxg),
                        theta=slice(jys22 + 1 - myg, jys22 + 1),
                    ).values,
                    v_lower_outer_intersep.isel(theta=slice(myg)).values,
                )

            v_lower_outer_SOL = v.bout.from_region("lower_outer_SOL")

            # Remove attributes that are expected to be different
            del v_lower_outer_SOL.attrs["regions"]
            xrt.assert_identical(
                v_noregions.isel(
                    x=slice(ixs2 - mxg, None), theta=slice(jys22 + 1, None)
                ),
                v_lower_outer_SOL.isel(theta=slice(myg, None)),
            )
            if myg > 0:
                # check y-guards, which were 'communicated' by from_region
                # Coordinates are not equal, so only compare array values
                npt.assert_equal(
                    v_noregions.isel(
                        x=slice(ixs2 - mxg, None),
                        theta=slice(jys22 + 1 - myg, jys22 + 1),
                    ).values,
                    v_lower_outer_SOL.isel(theta=slice(myg)).values,
                )

    def test_interpolate_parallel_all_variables_arg(self, bout_xyt_example_files):
        # Check that passing 'variables=...' to interpolate_parallel() does actually
        # interpolate all the variables
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=1,
            nype=1,
            nt=1,
            grid="grid",
            topology="sol",
        )

        ds = open_boutdataset(
            datapath=dataset_list, gridfilepath=grid_ds, geometry="toroidal"
        )

        # Get high parallel resolution version of ds, and check that
        ds = ds.bout.interpolate_parallel(...)

        interpolated_variables = [v for v in ds]

        assert set(interpolated_variables) == set(
            (
                "n",
                "T",
                "S",
                "g11",
                "g22",
                "g33",
                "g12",
                "g13",
                "g23",
                "g_11",
                "g_22",
                "g_33",
                "g_12",
                "g_13",
                "g_23",
                "G1",
                "G2",
                "G3",
                "J",
                "Bxy",
            )
        )

    def test_interpolate_parallel_limiter(
        self,
        bout_xyt_example_files,
    ):
        # This test checks that the regions created in the new high-resolution Dataset by
        # interpolate_parallel are correct.
        # This test does not test the accuracy of the parallel interpolation (there are
        # other tests for that).
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            guards={"x": 2, "y": 2},
            grid="grid",
            topology="limiter",
        )
        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=True,
            keep_yboundaries=False,
        )
        # Get high parallel resolution version of ds, and check that
        ds = ds.bout.interpolate_parallel(["n", "T"])
        mxg = 2
        myg = 2
        ixs1 = ds.metadata["ixseps1"]
        for var in ["n", "T"]:
            v = ds[var]
            v_noregions = v.copy(deep=True)
            # Remove attributes that are expected to be different
            del v_noregions.attrs["regions"]
            v_core = v.bout.from_region("core", with_guards={"theta": 0})
            # Remove attributes that are expected to be different
            del v_core.attrs["regions"]
            xrt.assert_identical(v_noregions.isel(x=slice(ixs1 + mxg)), v_core)
            v_sol = v.bout.from_region("SOL")
            # Remove attributes that are expected to be different
            del v_sol.attrs["regions"]
            xrt.assert_identical(v_noregions.isel(x=slice(ixs1 - mxg, None)), v_sol)

    def test_integrate_midpoints_slab(self, bout_xyt_example_files):
        # Create data
        dataset_list = bout_xyt_example_files(
            None, lengths=(4, 100, 110, 120), nxpe=1, nype=1, nt=1, syn_data_type=1
        )
        ds = open_boutdataset(dataset_list)
        t = np.linspace(0.0, 8.0, 4)[:, np.newaxis, np.newaxis, np.newaxis]
        x = np.linspace(0.05, 9.95, 100)[np.newaxis, :, np.newaxis, np.newaxis]
        y = np.linspace(0.1, 21.9, 110)[np.newaxis, np.newaxis, :, np.newaxis]
        z = np.linspace(0.15, 35.85, 120)[np.newaxis, np.newaxis, np.newaxis, :]
        ds["t"].data[...] = t.squeeze()
        ds["dx"].data[...] = 0.1
        ds["dy"].data[...] = 0.2
        ds["dz"] = 0.3
        tfunc = 1.5 * t
        xfunc = x**2
        yfunc = 10.0 * y - y**2
        zfunc = 2.0 * z**2 - 30.0 * z
        ds["n"].data[...] = tfunc * xfunc * yfunc * zfunc
        tintegral = 48.0
        xintegral = 1000.0 / 3.0
        yintegral = 5.0 * 22.0**2 - 22.0**3 / 3.0
        zintegral = 2.0 * 36.0**3 / 3.0 - 15.0 * 36.0**2
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims="t"),
            (tintegral * xfunc * yfunc * zfunc).squeeze(),
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims="x"),
            (tfunc * xintegral * yfunc * zfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims="y"),
            (tfunc * xfunc * yintegral * zfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims="z"),
            (tfunc * xfunc * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x"]),
            (tintegral * xintegral * yfunc * zfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "y"]),
            (tintegral * xfunc * yintegral * zfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "z"]),
            (tintegral * xfunc * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["x", "y"]),
            (tfunc * xintegral * yintegral * zfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["x", "z"]),
            (tfunc * xintegral * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["y", "z"]),
            (tfunc * xfunc * yintegral * zintegral).squeeze(),
            rtol=1.2e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x", "y"]),
            (tintegral * xintegral * yintegral * zfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x", "z"]),
            (tintegral * xintegral * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "y", "z"]),
            (tintegral * xfunc * yintegral * zintegral).squeeze(),
            rtol=1.2e-4,
        )
        # default dims
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n"),
            (tfunc * xintegral * yintegral * zintegral).squeeze(),
            rtol=1.4e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x", "y", "z"]),
            (tintegral * xintegral * yintegral * zintegral),
            rtol=1.4e-4,
        )

        # Create and test Field2D
        ds["S"][...] = (tfunc * xfunc * yfunc).squeeze()

        # S is 'axisymmetric' so z-integral is just the length of the
        # z-dimension
        zintegral = 36.0

        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims="t"),
            (tintegral * xfunc * yfunc).squeeze(),
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims="x"),
            (tfunc * xintegral * yfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims="y"),
            (tfunc * xfunc * yintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims="z"),
            (tfunc * xfunc * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "x"]),
            (tintegral * xintegral * yfunc).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "y"]),
            (tintegral * xfunc * yintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "z"]),
            (tintegral * xfunc * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["x", "y"]),
            (tfunc * xintegral * yintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["x", "z"]),
            (tfunc * xintegral * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["y", "z"]),
            (tfunc * xfunc * yintegral * zintegral).squeeze(),
            rtol=1.2e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "x", "y"]),
            (tintegral * xintegral * yintegral),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "x", "z"]),
            (tintegral * xintegral * yfunc * zintegral).squeeze(),
            rtol=1.0e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "y", "z"]),
            (tintegral * xfunc * yintegral * zintegral).squeeze(),
            rtol=1.2e-4,
        )
        # default dims
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S"),
            (tfunc * xintegral * yintegral * zintegral).squeeze(),
            rtol=1.4e-4,
        )
        npt.assert_allclose(
            ds.bout.integrate_midpoints("S", dims=["t", "x", "y", "z"]),
            (tintegral * xintegral * yintegral * zintegral),
            rtol=1.4e-4,
        )

    @pytest.mark.parametrize(
        "location", ["CELL_CENTRE", "CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"]
    )
    def test_integrate_midpoints_salpha(self, bout_xyt_example_files, location):
        # Create data
        nx = 100
        ny = 110
        nz = 120
        dataset_list = bout_xyt_example_files(
            None,
            lengths=(4, nx, ny, nz),
            nxpe=1,
            nype=1,
            nt=1,
            syn_data_type=1,
            guards={"x": 2, "y": 2},
        )
        ds = open_boutdataset(dataset_list)

        ds, options = set_geometry_from_input_file(ds, "s-alpha.inp")

        ds = apply_geometry(ds, "toroidal")

        # Integrate 1 so we just get volume, areas and lengths
        ds["n"].values[:] = 1.0
        ds["n"].attrs["cell_location"] = location

        # remove boundary cells (don't want to integrate over those)
        ds = ds.bout.remove_yboundaries()
        if ds.metadata["keep_xboundaries"] and ds.metadata["MXG"] > 0:
            mxg = ds.metadata["MXG"]
            xslice = slice(mxg, -mxg)
        else:
            xslice = slice(None)
        ds = ds.isel(x=xslice)

        # Test geometry has major radius R and goes between minor radii a-Lr/2 and
        # a+Lr/2
        R = options.evaluate_scalar("mesh:R0")
        a = options.evaluate_scalar("mesh:a")
        Lr = options.evaluate_scalar("mesh:Lr")
        rinner = a - Lr / 2.0
        router = a + Lr / 2.0
        r = options.evaluate("mesh:r").squeeze()[xslice]
        if location == "CELL_XLOW":
            rinner = rinner - Lr / (2.0 * nx)
            router = router - Lr / (2.0 * nx)
            r = r - Lr / (2.0 * nx)
        q = options.evaluate_scalar("mesh:q")
        T_total = (ds.sizes["t"] - 1) * (ds["t"][1] - ds["t"][0]).values
        T_cumulative = np.arange(ds.sizes["t"]) * (ds["t"][1] - ds["t"][0]).values

        # Volume of torus with circular cross-section of major radius R and minor radius
        # a is 2*pi*R*pi*a^2
        # https://en.wikipedia.org/wiki/Torus
        # Default is to integrate over all spatial dimensions
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n"),
            2.0 * np.pi * R * np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Pass all spatial dims explicitly
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["x", "theta", "zeta"]),
            2.0 * np.pi * R * np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x", "theta", "zeta"]),
            T_total * 2.0 * np.pi * R * np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time using dims=...
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=...),
            T_total * 2.0 * np.pi * R * np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints(
                "n", dims=["t", "x", "theta", "zeta"], cumulative_t=True
            ),
            T_cumulative * 2.0 * np.pi * R * np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )

        # Area of torus with circular cross-section of major radius R and minor radius a
        # is 2*pi*R*2*pi*a
        # https://en.wikipedia.org/wiki/Torus
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["theta", "zeta"]),
            (2.0 * np.pi * R * 2.0 * np.pi * r)[np.newaxis, :]
            * np.ones(ds.sizes["t"])[:, np.newaxis],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "theta", "zeta"]),
            T_total * 2.0 * np.pi * R * 2.0 * np.pi * r,
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints(
                "n", dims=["t", "theta", "zeta"], cumulative_t=True
            ),
            T_cumulative[:, np.newaxis]
            * (2.0 * np.pi * R * 2.0 * np.pi * r)[np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )

        # Area of cross section in poloidal plane is difference of circle with radius
        # router and circle with radius rinner, pi*(router**2 - rinner**2)
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["x", "theta"]),
            np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x", "theta"]),
            T_total * np.pi * (router**2 - rinner**2),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints(
                "n", dims=["t", "x", "theta"], cumulative_t=True
            ),
            T_cumulative[:, np.newaxis]
            * np.pi
            * (router**2 - rinner**2)
            * np.ones(ds.sizes["zeta"])[np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )

        # x-z planes are 'conical frustrums', with area pi*(Rinner + Router)*Lr
        # https://en.wikipedia.org/wiki/Frustum
        theta = _1d_coord_from_spacing(ds["dy"], "theta").values
        if location == "CELL_YLOW":
            theta = theta - 2.0 * np.pi / (2.0 * ny)
        Rinner = R + rinner * np.cos(theta)
        Router = R + router * np.cos(theta)
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["x", "zeta"]),
            (np.pi * (Rinner + Router) * Lr)[np.newaxis, :]
            * np.ones(ds.sizes["t"])[:, np.newaxis],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x", "zeta"]),
            T_total * (np.pi * (Rinner + Router) * Lr),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints(
                "n", dims=["t", "x", "zeta"], cumulative_t=True
            ),
            T_cumulative[:, np.newaxis]
            * (np.pi * (Rinner + Router) * Lr)
            * np.ones(ds.sizes["theta"])[np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )

        # Radial lines have length Lr
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["x"]),
            Lr,
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x"]),
            T_total * Lr,
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "x"], cumulative_t=True),
            T_cumulative[:, np.newaxis, np.newaxis]
            * Lr
            * np.ones([ds.sizes["theta"], ds.sizes["zeta"]])[np.newaxis, :, :],
            rtol=1.0e-5,
            atol=0.0,
        )

        # Poloidal lines have length 2*pi*r
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["theta"]),
            (2.0 * np.pi * r)[np.newaxis, :, np.newaxis]
            * np.ones(ds.sizes["t"])[:, np.newaxis, np.newaxis]
            * np.ones(ds.sizes["zeta"])[np.newaxis, np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "theta"]),
            T_total
            * (2.0 * np.pi * r)[:, np.newaxis]
            * np.ones(ds.sizes["zeta"])[np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "theta"], cumulative_t=True),
            T_cumulative[:, np.newaxis, np.newaxis]
            * (2.0 * np.pi * r)[np.newaxis, :, np.newaxis]
            * np.ones(ds.sizes["zeta"])[np.newaxis, np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )

        # Field line length
        # -----------------
        #
        # Field lines, parameterised by poloidal angle theta, have (R, Z, zeta)
        # coordinates
        # X = (R + r*cos(theta),
        #      -r*sin(theta),
        #      -q*2*atan(sqrt((1-r/R0)/(1+r/R0))*tan(theta/2))
        #     )
        # using d(arctan(x))/dx = 1/(1 + x**2) and d(tan(x))/dx = 1/cos(x)**2
        # dX/dtheta = (r*sin(theta),
        #              + r*cos(theta),
        #              - q * 2
        #                * sqrt((1-r/R0)/(1+r/R0))/(2*cos(theta/2)**2)
        #                / (1 + (1-r/R0)/(1+r/R0)*tan(theta/2)**2)
        #             )
        #           = (r*sin(theta),
        #              + r*cos(theta),
        #              - q
        #                * sqrt((1-r/R0)/(1+r/R0))/cos(theta/2)**2
        #                / (1 + (1-r/R0)/(1+r/R0)*tan(theta/2)**2)
        #             )
        #           = (r*sin(theta),
        #              + r*cos(theta),
        #              - q
        #                * sqrt((1-r/R0)/(1+r/R0))
        #                / (cos(theta/2)**2 + (1-r/R0)/(1+r/R0)*sin(theta/2)**2)
        #             )
        # Line element dl = |dR + dZ + R dzeta|
        #                 = |dR + dZ + (R0 + r cos(theta)) dzeta|
        #                 = |dX/dtheta|*dtheta
        # |dX/dtheta| = |
        #             = sqrt(
        #                r**2*sin(theta)**2
        #                + r**2*cos(theta)**2
        #                + (R0 + r cos(theta))**2 * q**2
        #                  * (1-r/R0)/(1+r/R0)
        #                  / (cos(theta/2)**2 + (1-r/R0)/(1+r/R0)*sin(theta/2)**2)**2
        #               )
        #             = sqrt(
        #                r**2
        #                + (R0 + r cos(theta))**2 * q**2
        #                  * (1-r/R0)/(1+r/R0)
        #                  / (cos(theta/2)**2 + (1-r/R0)/(1+r/R0)*sin(theta/2)**2)**2
        #               )
        # field line length is int_{0}^{2 pi} dl
        a = r**2
        c = (1 - r / R) / (1 + r / R)
        b = q**2 * c

        def func(theta):
            return np.sqrt(
                a
                + b
                * (R + r * np.cos(theta)) ** 2
                / (np.cos(theta / 2.0) ** 2 + c * np.sin(theta / 2.0) ** 2) ** 2
            )

        integral, _ = quad_vec(func, 0.0, 2.0 * np.pi)

        ds["n_aligned"] = ds["n"].bout.to_field_aligned()
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n_aligned", dims=["theta"]),
            integral[np.newaxis, :, np.newaxis]
            * np.ones(ds.sizes["t"])[:, np.newaxis, np.newaxis]
            * np.ones(ds.sizes["zeta"])[np.newaxis, np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n_aligned", dims=["t", "theta"]),
            T_total
            * integral[:, np.newaxis]
            * np.ones(ds.sizes["zeta"])[np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints(
                "n_aligned", dims=["t", "theta"], cumulative_t=True
            ),
            T_cumulative[:, np.newaxis, np.newaxis]
            * integral[np.newaxis, :, np.newaxis]
            * np.ones(ds.sizes["zeta"])[np.newaxis, np.newaxis, :],
            rtol=1.0e-5,
            atol=0.0,
        )

        # Toroidal lines have length 2*pi*Rxy
        if location == "CELL_CENTRE":
            R_2d = ds["R"]
        else:
            R_2d = ds[f"Rxy_{location}"]
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["zeta"]),
            (2.0 * np.pi * R_2d).values[np.newaxis, :, :]
            * np.ones(ds.sizes["t"])[:, np.newaxis, np.newaxis],
            rtol=1.0e-5,
            atol=0.0,
        )
        # Integrate in time too
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "zeta"]),
            T_total * (2.0 * np.pi * R_2d),
            rtol=1.0e-5,
            atol=0.0,
        )
        # Cumulative integral in time
        npt.assert_allclose(
            ds.bout.integrate_midpoints("n", dims=["t", "zeta"], cumulative_t=True),
            T_cumulative[:, np.newaxis, np.newaxis]
            * (2.0 * np.pi * R_2d).values[np.newaxis, :, :],
            rtol=1.0e-5,
            atol=0.0,
        )

    def test_interpolate_from_unstructured(self, bout_xyt_example_files):
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            grid="grid",
            topology="upper-disconnected-double-null",
        )

        ds = open_boutdataset(
            datapath=dataset_list, gridfilepath=grid_ds, geometry="toroidal"
        )

        # Set up non-trivial R and Z coordinates so we have something to interpolate to
        r = ds["x"] / ds["x"][-1] * 0.1 + 0.1
        theta = ds["theta"] / ds["theta"][-1] * 2.0 * np.pi

        ds["R"] = r * np.cos(theta)
        ds["Z"] = 2.0 * r * np.sin(theta)

        ds["n"] = (ds["R"] + ds["Z"]).broadcast_like(ds["n"])

        n = ds["n"]

        NR = 23
        R_rect = np.linspace(-0.1, 0.1, NR)

        NZ = 37
        Z_rect = np.linspace(-0.2, 0.2, NZ)

        n_rect = n.bout.interpolate_from_unstructured(R=R_rect, Z=Z_rect)

        # check BoutDataset and BoutDataArray versions are consistent
        n_rect_from_ds = ds.bout.interpolate_from_unstructured(..., R=R_rect, Z=Z_rect)[
            "n"
        ]
        npt.assert_allclose(n_rect, n_rect_from_ds)

        # check non-NaN values are correct
        n_check = R_rect[:, np.newaxis] + Z_rect[np.newaxis, :]
        n_rect = n_rect.isel(t=0, zeta=0).transpose("R", "Z")
        mask = ~np.isnan(n_rect.values)
        npt.assert_allclose(n_rect.values[mask], n_check[mask], atol=1.0e-7)

        # Check there were non-nan values to compare in the previous assert_allclose
        assert int((~np.isnan(n_rect)).count()) == 851

    def test_interpolate_from_unstructured_unstructured_output(
        self, bout_xyt_example_files
    ):
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            grid="grid",
            topology="lower-disconnected-double-null",
        )

        ds = open_boutdataset(
            datapath=dataset_list, gridfilepath=grid_ds, geometry="toroidal"
        )

        # Set up non-trivial R and Z coordinates so we have something to interpolate to
        r = ds["x"] / ds["x"][-1] * 0.1 + 0.1
        theta = ds["theta"] / ds["theta"][-1] * 2.0 * np.pi

        ds["R"] = r * np.cos(theta)
        ds["Z"] = 2.0 * r * np.sin(theta)

        ds["n"] = (ds["R"] + ds["Z"]).broadcast_like(ds["n"])

        n = ds["n"]

        # make a set of points within the domain
        # domain is 'double null' so leave a gap to avoid the 'upper target' boundaries
        t = np.concatenate(
            [
                np.linspace(0.2, np.pi - 0.2, 11),
                np.linspace(np.pi + 0.2, 2.0 * np.pi - 0.2, 13),
            ]
        )
        R_unstruct = (0.142679 + 0.013291 * t / (2.0 * np.pi)) * np.cos(t)
        Z_unstruct = (0.240837 + 0.113408 * t / (2.0 * np.pi)) * np.sin(t)

        # check input validation
        with pytest.raises(ValueError):
            # input length mismatch
            n_unstruct = n.bout.interpolate_from_unstructured(
                R=R_unstruct, Z=Z_unstruct[:-2], structured_output=False
            )
        with pytest.raises(ValueError):
            # input length mismatch
            n_unstruct = n.bout.interpolate_from_unstructured(
                R=R_unstruct[:-2], Z=Z_unstruct, structured_output=False
            )

        n_unstruct = n.bout.interpolate_from_unstructured(
            R=R_unstruct, Z=Z_unstruct, structured_output=False
        )

        # check BoutDataset and BoutDataArray versions are consistent
        n_unstruct_from_ds = ds.bout.interpolate_from_unstructured(
            ..., R=R_unstruct, Z=Z_unstruct, structured_output=False
        )["n"]
        npt.assert_allclose(n_unstruct, n_unstruct_from_ds)

        # check non-NaN values are correct
        n_check = R_unstruct + Z_unstruct
        n_unstruct = n_unstruct.isel(t=0, zeta=0)
        npt.assert_allclose(n_unstruct.values, n_check, atol=1.0e-7)

    def test_interpolate_to_cartesian(self, bout_xyt_example_files):
        dataset_list = bout_xyt_example_files(
            None, lengths=(2, 16, 17, 18), nxpe=1, nype=1, nt=1
        )
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["g11"].copy(deep=True)
        ds["Rxy"] = ds["g11"].copy(deep=True)
        ds["Zxy"] = ds["g11"].copy(deep=True)

        r = np.linspace(1.0, 2.0, ds.metadata["nx"])
        theta = np.linspace(0.0, 2.0 * np.pi, ds.metadata["ny"])
        R = r[:, np.newaxis] * np.cos(theta[np.newaxis, :])
        Z = r[:, np.newaxis] * np.sin(theta[np.newaxis, :])
        ds["Rxy"].values[:] = R
        ds["Zxy"].values[:] = Z

        ds = apply_geometry(ds, "toroidal")

        ds["n"].values[:] = 1.0

        nX = 30
        nY = 30
        nZ = 10
        ds_cartesian = ds.bout.interpolate_to_cartesian(nX, nY, nZ)

        # Check a point inside the original grid
        npt.assert_allclose(
            ds_cartesian["n"]
            .isel(t=0, X=round(nX * 4 / 5), Y=nY // 2, Z=nZ // 2)
            .item(),
            1.0,
            rtol=1.0e-15,
            atol=1.0e-15,
        )
        # Check a point outside the original grid
        assert np.isnan(ds_cartesian["n"].isel(t=0, X=0, Y=0, Z=0).item())
        # Check output is float32
        assert ds_cartesian["n"].dtype == np.float32

    def test_add_cartesian_coordinates(self, bout_xyt_example_files):
        dataset_list = bout_xyt_example_files(None, nxpe=1, nype=1, nt=1)
        with pytest.warns(UserWarning):
            ds = open_boutdataset(
                datapath=dataset_list, inputfilepath=None, keep_xboundaries=False
            )

        ds["psixy"] = ds["g11"].copy(deep=True)
        ds["Rxy"] = ds["g11"].copy(deep=True)
        ds["Zxy"] = ds["g11"].copy(deep=True)

        R0 = 3.0
        r = np.linspace(1.0, 2.0, ds.metadata["nx"])
        theta = np.linspace(0.0, 2.0 * np.pi, ds.metadata["ny"])
        R = R0 + r[:, np.newaxis] * np.cos(theta[np.newaxis, :])
        Z = r[:, np.newaxis] * np.sin(theta[np.newaxis, :])
        ds["Rxy"].values[:] = R
        ds["Zxy"].values[:] = Z

        ds = apply_geometry(ds, "toroidal")

        zeta = ds["zeta"].values

        ds = ds.bout.add_cartesian_coordinates()

        npt.assert_allclose(
            ds["X_cartesian"],
            R[:, :, np.newaxis] * np.cos(zeta[np.newaxis, np.newaxis, :]),
        )
        npt.assert_allclose(
            ds["Y_cartesian"],
            R[:, :, np.newaxis] * np.sin(zeta[np.newaxis, np.newaxis, :]),
        )
        npt.assert_allclose(
            ds["Z_cartesian"],
            Z[:, :, np.newaxis] * np.ones(ds.metadata["nz"])[np.newaxis, np.newaxis, :],
        )


class TestLoadInputFile:
    @pytest.mark.skip
    def test_load_options(self):
        from boutdata.data import BoutOptionsFile, BoutOptions

        options = BoutOptionsFile(EXAMPLE_OPTIONS_FILE_PATH)
        assert isinstance(options, BoutOptions)
        # TODO Check it contains the same text

    @pytest.mark.skip
    def test_load_options_in_dataset(self, bout_xyt_example_files):
        dataset_list = bout_xyt_example_files(None, nxpe=1, nype=1, nt=1)
        ds = open_boutdataset(
            datapath=dataset_list, inputfilepath=EXAMPLE_OPTIONS_FILE_PATH
        )
        assert isinstance(ds.options, BoutOptions)


@pytest.mark.skip(reason="Not yet implemented")
class TestLoadLogFile:
    pass


class TestSave:
    def test_save_all(self, tmp_path_factory, bout_xyt_example_files):
        # Create data
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=5, nt=1, write_to_disk=True
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            original = open_boutdataset(datapath=path, inputfilepath=None)

        # Save it to a netCDF file
        savedir = tmp_path_factory.mktemp("test_save_all")
        savepath = savedir.joinpath("temp_boutdata.nc")
        original.bout.save(savepath=savepath)

        # Load it again using bare xarray
        recovered = open_dataset(savepath)

        # Compare equal (not identical because attributes are changed when saving)
        xrt.assert_equal(original, recovered)

    @pytest.mark.parametrize("geometry", [None, "toroidal"])
    def test_reload_all(self, tmp_path_factory, bout_xyt_example_files, geometry):
        # Create data
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=5, nt=1, grid="grid", write_to_disk=True
        )

        gridpath = path.parent.joinpath("grid.nc")

        # Load it as a boutdataset
        if geometry is None:
            with pytest.warns(UserWarning):
                original = open_boutdataset(
                    datapath=path,
                    inputfilepath=None,
                    geometry=geometry,
                    gridfilepath=None if geometry is None else gridpath,
                )
        else:
            original = open_boutdataset(
                datapath=path,
                inputfilepath=None,
                geometry=geometry,
                gridfilepath=None if geometry is None else gridpath,
            )

        # Save it to a netCDF file
        savedir = tmp_path_factory.mktemp("test_reload_all")
        savepath = savedir.joinpath("temp_boutdata.nc")
        original.bout.save(savepath=savepath)

        # Load it again
        recovered = open_boutdataset(savepath)

        xrt.assert_identical(original.load(), recovered.load())

        # Check if we can load with a different geometry argument
        for reload_geometry in [None, "toroidal"]:
            if reload_geometry is None or geometry == reload_geometry:
                recovered = open_boutdataset(
                    savepath,
                    geometry=reload_geometry,
                    gridfilepath=None if reload_geometry is None else gridpath,
                )
                xrt.assert_identical(original.load(), recovered.load())
            else:
                # Expect a warning because we change the geometry
                print("here", gridpath)
                with pytest.warns(UserWarning):
                    recovered = open_boutdataset(
                        savepath, geometry=reload_geometry, gridfilepath=gridpath
                    )
                # Datasets won't be exactly the same because different geometry was
                # applied

    @pytest.mark.parametrize("save_dtype", [np.float64, np.float32])
    @pytest.mark.parametrize(
        "separate_vars", [False, pytest.param(True, marks=pytest.mark.long)]
    )
    def test_save_dtype(
        self, tmp_path_factory, bout_xyt_example_files, save_dtype, separate_vars
    ):

        # Create data
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=1, nype=1, nt=1, write_to_disk=True
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            original = open_boutdataset(datapath=path, inputfilepath=None)

        # Save it to a netCDF file
        savedir = tmp_path_factory.mktemp("test_save_dtype")
        savepath = savedir.joinpath("temp_boutdata.nc")
        original.bout.save(
            savepath=savepath, save_dtype=save_dtype, separate_vars=separate_vars
        )

        # Load it again using bare xarray
        if separate_vars:
            for v in ["n", "T"]:
                savepath = savedir.joinpath(f"temp_boutdata_{v}.nc")
                recovered = open_dataset(savepath)
                assert recovered[v].values.dtype == np.dtype(save_dtype)
        else:
            recovered = open_dataset(savepath)

            for v in original:
                assert recovered[v].values.dtype == np.dtype(save_dtype)

    def test_save_separate_variables(self, tmp_path_factory, bout_xyt_example_files):
        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=1, nt=1, write_to_disk=True
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            original = open_boutdataset(datapath=path, inputfilepath=None)

        # Save it to a netCDF file
        savedir = tmp_path_factory.mktemp("test_save_separate_variables")
        savepath = savedir.joinpath("temp_boutdata.nc")
        original.bout.save(savepath=savepath, separate_vars=True)

        for var in ["n", "T"]:
            # Load it again using bare xarray
            savepath = savedir.joinpath(f"temp_boutdata_{var}.nc")
            recovered = open_dataset(savepath)

            # Compare equal (not identical because attributes are changed when saving)
            xrt.assert_equal(recovered[var], original[var])

        # test open_boutdataset() on dataset saved with separate_vars=True
        savepath = savedir.joinpath("temp_boutdata_*.nc")
        recovered = open_boutdataset(savepath)
        xrt.assert_identical(original, recovered)

    @pytest.mark.parametrize("geometry", [None, "toroidal"])
    def test_reload_separate_variables(
        self, tmp_path_factory, bout_xyt_example_files, geometry
    ):
        if geometry is not None:
            grid = "grid"
        else:
            grid = None

        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=1, nt=1, grid=grid, write_to_disk=True
        )

        if grid is not None:
            gridpath = path.parent.joinpath("grid.nc")
        else:
            gridpath = None

        # Load it as a boutdataset
        if geometry is None:
            with pytest.warns(UserWarning):
                original = open_boutdataset(
                    datapath=path,
                    inputfilepath=None,
                    geometry=geometry,
                    gridfilepath=gridpath,
                )
        else:
            original = open_boutdataset(
                datapath=path,
                inputfilepath=None,
                geometry=geometry,
                gridfilepath=gridpath,
            )

        # Save it to a netCDF file
        savedir = tmp_path_factory.mktemp("test_reload_separate_variables")
        savepath = savedir.joinpath("temp_boutdata.nc")
        original.bout.save(savepath=savepath, separate_vars=True)

        # Load it again
        savepath = savedir.joinpath("temp_boutdata_*.nc")
        recovered = open_boutdataset(savepath)

        # Compare
        xrt.assert_identical(recovered, original)

    @pytest.mark.parametrize("geometry", [None, "toroidal"])
    def test_reload_separate_variables_time_split(
        self, tmp_path_factory, bout_xyt_example_files, geometry
    ):
        if geometry is not None:
            grid = "grid"
        else:
            grid = None

        path = bout_xyt_example_files(
            tmp_path_factory, nxpe=4, nype=1, nt=1, grid=grid, write_to_disk=True
        )

        if grid is not None:
            gridpath = path.parent.joinpath("grid.nc")
        else:
            gridpath = None

        # Load it as a boutdataset
        if geometry is None:
            with pytest.warns(UserWarning):
                original = open_boutdataset(
                    datapath=path,
                    inputfilepath=None,
                    geometry=geometry,
                    gridfilepath=gridpath,
                )
        else:
            original = open_boutdataset(
                datapath=path,
                inputfilepath=None,
                geometry=geometry,
                gridfilepath=gridpath,
            )

        # Save it to a netCDF file
        tcoord = original.metadata.get("bout_tdim", "t")
        savedir = tmp_path_factory.mktemp("test_reload_separate_variables_time_split")
        savepath = savedir.joinpath("temp_boutdata_1.nc")
        original.isel({tcoord: slice(3)}).bout.save(
            savepath=savepath, separate_vars=True
        )
        savepath = savedir.joinpath("temp_boutdata_2.nc")
        original.isel({tcoord: slice(3, None)}).bout.save(
            savepath=savepath, separate_vars=True
        )

        # Load it again
        savepath = savedir.joinpath("temp_boutdata_*.nc")
        recovered = open_boutdataset(savepath)

        # Compare
        xrt.assert_identical(recovered, original)


class TestSaveRestart:
    @pytest.mark.parametrize("tind", [None, pytest.param(1, marks=pytest.mark.long)])
    def test_to_restart(self, tmp_path_factory, bout_xyt_example_files, tind):
        nxpe = 3
        nype = 2

        nt = 6

        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=nxpe,
            nype=nype,
            nt=1,
            lengths=[nt, 4, 4, 7],
            guards={"x": 2, "y": 2},
            write_to_disk=True,
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(datapath=path)

        nx = ds.metadata["nx"]
        ny = ds.metadata["ny"]

        # Save it to a netCDF file
        savepath = tmp_path_factory.mktemp("test_to_restart")
        if tind is None:
            ds.bout.to_restart(savepath=savepath, nxpe=nxpe, nype=nype)
        else:
            ds.bout.to_restart(savepath=savepath, nxpe=nxpe, nype=nype, tind=tind)

        mxsub = (nx - 4) // nxpe
        mysub = ny // nype
        for proc_yind in range(nype):
            for proc_xind in range(nxpe):
                num = nxpe * proc_yind + proc_xind

                restart_ds = open_dataset(savepath.joinpath(f"BOUT.restart.{num}.nc"))

                if tind is None:
                    t = -1
                else:
                    t = tind

                # ignore guard cells - they are filled with NaN in the created restart
                # files
                restart_ds = restart_ds.isel(x=slice(2, -2), y=slice(2, -2))

                check_ds = ds.isel(
                    t=t,
                    x=slice(2 + proc_xind * mxsub, 2 + (proc_xind + 1) * mxsub),
                    y=slice(proc_yind * mysub, (proc_yind + 1) * mysub),
                ).load()
                t_array = check_ds["t"]
                check_ds = check_ds.drop_vars(["t", "x", "y", "z"])

                # No coordinates saved in restart files, so unset them in check_ds
                check_ds = check_ds.reset_coords()

                # ignore variables that depend on the rank of the BOUT++ process - they
                # cannot be consistent with check_ds
                rank_dependent_vars = ["PE_XIND", "PE_YIND", "MYPE"]
                for v in restart_ds:
                    if v in check_ds:
                        xrt.assert_equal(restart_ds[v], check_ds[v])
                    else:
                        if v == "hist_hi":
                            if t >= 0:
                                assert restart_ds[v].values == t
                            else:
                                assert restart_ds[v].values == nt - 1
                        elif v == "tt":
                            assert restart_ds[v].values == t_array
                        elif v not in rank_dependent_vars:
                            assert restart_ds[v].values == check_ds.metadata[v]

    def test_to_restart_change_npe(self, tmp_path_factory, bout_xyt_example_files):
        nxpe_in = 3
        nype_in = 2

        nxpe = 2
        nype = 4

        nt = 6

        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=nxpe_in,
            nype=nype_in,
            nt=1,
            lengths=[nt, 4, 4, 7],
            guards={"x": 2, "y": 2},
            write_to_disk=True,
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(datapath=path)

        nx = ds.metadata["nx"]
        ny = ds.metadata["ny"]

        # Save it to a netCDF file
        savepath = tmp_path_factory.mktemp("test_to_restart_change_npe")
        ds.bout.to_restart(savepath=savepath, nxpe=nxpe, nype=nype)

        mxsub = (nx - 4) // nxpe
        mysub = ny // nype
        for proc_yind in range(nype):
            for proc_xind in range(nxpe):
                num = nxpe * proc_yind + proc_xind

                restart_ds = open_dataset(savepath.joinpath(f"BOUT.restart.{num}.nc"))

                # ignore guard cells - they are filled with NaN in the created restart
                # files
                restart_ds = restart_ds.isel(x=slice(2, -2), y=slice(2, -2))

                check_ds = ds.isel(
                    t=-1,
                    x=slice(2 + proc_xind * mxsub, 2 + (proc_xind + 1) * mxsub),
                    y=slice(proc_yind * mysub, (proc_yind + 1) * mysub),
                ).load()
                t_array = check_ds["t"]
                check_ds = check_ds.drop_vars(["t", "x", "y", "z"])

                # No coordinates saved in restart files, so unset them in check_ds
                check_ds = check_ds.reset_coords()

                # ignore variables that depend on the rank of the BOUT++ process - they
                # cannot be consistent with check_ds
                rank_dependent_vars = ["PE_XIND", "PE_YIND", "MYPE"]
                for v in restart_ds:
                    if v in check_ds:
                        xrt.assert_equal(restart_ds[v], check_ds[v])
                    else:
                        if v in ["NXPE", "NYPE", "MXSUB", "MYSUB"]:
                            pass
                        elif v == "hist_hi":
                            assert restart_ds[v].values == nt - 1
                        elif v == "tt":
                            assert restart_ds[v].values == t_array
                        elif v not in rank_dependent_vars:
                            assert restart_ds[v].values == check_ds.metadata[v]

    @pytest.mark.long
    def test_to_restart_change_npe_doublenull(
        self, tmp_path_factory, bout_xyt_example_files
    ):
        nxpe_in = 3
        nype_in = 6

        nxpe = 1
        nype = 12

        nt = 6

        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=nxpe_in,
            nype=nype_in,
            nt=1,
            guards={"x": 2, "y": 2},
            lengths=(nt, 5, 4, 7),
            topology="upper-disconnected-double-null",
            write_to_disk=True,
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(datapath=path)

        nx = ds.metadata["nx"]
        ny = ds.metadata["ny"]

        # Save it to a netCDF file
        savepath = tmp_path_factory.mktemp("test_to_restart_change_npe_doublenull")
        ds.bout.to_restart(savepath=savepath, nxpe=nxpe, nype=nype)

        mxsub = (nx - 4) // nxpe
        mysub = ny // nype
        for proc_yind in range(nype):
            for proc_xind in range(nxpe):
                num = nxpe * proc_yind + proc_xind

                restart_ds = open_dataset(savepath.joinpath(f"BOUT.restart.{num}.nc"))

                # ignore guard cells - they are filled with NaN in the created restart
                # files
                restart_ds = restart_ds.isel(x=slice(2, -2), y=slice(2, -2))

                check_ds = ds.isel(
                    t=-1,
                    x=slice(2 + proc_xind * mxsub, 2 + (proc_xind + 1) * mxsub),
                    y=slice(proc_yind * mysub, (proc_yind + 1) * mysub),
                ).load()
                t_array = check_ds["t"]
                check_ds = check_ds.drop_vars(["t", "x", "y", "z"])

                # No coordinates saved in restart files, so unset them in check_ds
                check_ds = check_ds.reset_coords()

                # ignore variables that depend on the rank of the BOUT++ process - they
                # cannot be consistent with check_ds
                rank_dependent_vars = ["PE_XIND", "PE_YIND", "MYPE"]
                for v in restart_ds:
                    if v in check_ds:
                        xrt.assert_equal(restart_ds[v], check_ds[v])
                    else:
                        if v in ["NXPE", "NYPE", "MXSUB", "MYSUB"]:
                            pass
                        elif v == "hist_hi":
                            assert restart_ds[v].values == nt - 1
                        elif v == "tt":
                            assert restart_ds[v].values == t_array
                        elif v not in rank_dependent_vars:
                            assert restart_ds[v].values == check_ds.metadata[v]

    @pytest.mark.long
    @pytest.mark.parametrize("npes", [(2, 6), (3, 4)])
    def test_to_restart_change_npe_doublenull_expect_fail(
        self, tmp_path_factory, bout_xyt_example_files, npes
    ):
        nxpe_in = 3
        nype_in = 6

        nxpe, nype = npes

        path = bout_xyt_example_files(
            tmp_path_factory,
            nxpe=nxpe_in,
            nype=nype_in,
            nt=1,
            guards={"x": 2, "y": 2},
            lengths=(6, 5, 4, 7),
            topology="lower-disconnected-double-null",
            write_to_disk=True,
        )

        # Load it as a boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(datapath=path)

        nx = ds.metadata["nx"]
        ny = ds.metadata["ny"]

        # Save it to a netCDF file
        savepath = tmp_path_factory.mktemp(
            "test_to_restart_change_npe_doublenull_expect_fail"
        )
        with pytest.raises(ValueError):
            ds.bout.to_restart(savepath=savepath, nxpe=nxpe, nype=nype)

    def test_from_restart_to_restart(self, tmp_path):
        datapath = Path(__file__).parent.joinpath(
            "data", "restart", "BOUT.restart.*.nc"
        )
        ds = open_boutdataset(datapath, keep_xboundaries=True, keep_yboundaries=True)

        ds.bout.to_restart(savepath=tmp_path, nxpe=1, nype=4)



"""
This example is to compare the speed of using dask + xarray vs the old collect function.

It just operates on the bout files in the local directory.
It first opens all the data then saves it again at a lower resolution.
"""

from timeit import Timer

from xcollect.boutdataset import open_boutdataset
from combine import combine_boutdata


def new_collect():
    ds = open_boutdataset(datapath="./BOUT.dmp.*.nc")
    ds.bout.save(save_dtype="float32", separate_vars=True)


def old_collect():
    combine_boutdata(
        datadir=".",
        prefix="/BOUT.dmp.",
        inpfile="BOUT.inp",
        output_file_prefix="boutdata_old",
        save_fields_separately=True,
        ignore_fields=["logn", "logT", "uE"],
        only_save_fields=None,
        save_grid=False,
        save_dtype="float32",
    )


t_new = Timer(new_collect).timeit(number=1)
t_old = Timer(old_collect).timeit(number=1)

print("New collect took " + t_new)
print("Old collect took " + t_old)



from copy import copy
from warnings import warn
from pathlib import Path
from functools import partial
from itertools import chain

from boutdata.data import BoutOptionsFile
import xarray as xr
from numpy import unique

from natsort import natsorted

from . import geometries
from .utils import _set_attrs_on_all_vars, _separate_metadata, _check_filetype, _is_path


_BOUT_PER_PROC_VARIABLES = [
    "wall_time",
    "wtime",
    "wtime_rhs",
    "wtime_invert",
    "wtime_comms",
    "wtime_io",
    "wtime_per_rhs",
    "wtime_per_rhs_e",
    "wtime_per_rhs_i",
    "PE_XIND",
    "PE_YIND",
    "MYPE",
]
_BOUT_TIME_DEPENDENT_META_VARS = ["iteration", "hist_hi", "tt"]


# This code should run whenever any function from this module is imported
# Set all attrs to survive all mathematical operations
# (see https://github.com/pydata/xarray/pull/2482)
try:
    xr.set_options(keep_attrs=True)
except ValueError:
    raise ImportError(
        "For dataset attributes to be permanent you need to be "
        "using the development version of xarray - found at "
        "https://github.com/pydata/xarray/"
    )
try:
    xr.set_options(file_cache_maxsize=256)
except ValueError:
    raise ImportError(
        "For open and closing of netCDF files correctly you need"
        " to be using the development version of xarray - found"
        " at https://github.com/pydata/xarray/"
    )


# TODO somehow check that we have access to the latest version of auto_combine


def open_boutdataset(
    datapath="./BOUT.dmp.*.nc",
    inputfilepath=None,
    geometry=None,
    gridfilepath=None,
    chunks=None,
    keep_xboundaries=True,
    keep_yboundaries=False,
    run_name=None,
    info=True,
    is_restart=None,
    **kwargs,
):
    """
    Load a dataset from a set of BOUT output files, including the input options
    file. Can also load from a grid file or from restart files.

    Note that when reloading a Dataset that was saved by xBOUT, the state of the saved
    Dataset is restored, and the values of `keep_xboundaries`, `keep_yboundaries`, and
    `run_name` are ignored. `geometry` is treated specially, and can be passed when
    reloading a Dataset (along with `gridfilepath` if needed).

    Troubleshooting
    ---------------
    Variable conflicts: sometimes, for example when loading data from multiple restarts,
    some variables may have conflicts (e.g. a source term was changed between some of
    the restarts, but the source term is saved as time-independent, without a
    t-dimension). In this case one workaround is to pass a list of variable names to the
    keyword argument `drop_vars` to ignore the variables with conflicts, e.g. if `"S1"`
    and `"S2"` have conflicts
    ```
    ds = open_boutdataset("data*/boutdata.nc", drop_variables=["S1", "S2"])
    ```
    will open a Dataset which is missing `"S1"` and `"S2"`.\
    [`drop_variables` is an argument of `xarray.open_dataset()` that is passed down
    through `kwargs`.]

    Parameters
    ----------
    datapath : str or (list or tuple of xr.Dataset), optional
        Path to the data to open. Can point to either a set of one or more dump
        files, or a single grid file.

        To specify multiple dump files you must enter the path to them as a
        single glob, e.g. './BOUT.dmp.*.nc', or for multiple consecutive runs
        in different directories (in order) then './run*/BOUT.dmp.*.nc'.

        If a list or tuple of xr.Dataset is passed, they will be combined with
        xr.combine_nested() instead of loading data from disk (intended for unit
        testing).
    chunks : dict, optional
    inputfilepath : str, optional
    geometry : str, optional
        The geometry type of the grid data. This will specify what type of
        coordinates to add to the dataset, e.g. 'toroidal' or 'cylindrical'.

        If not specified then will attempt to read it from the file attrs.
        If still not found then a warning will be thrown, which can be
        suppressed by passing `info`=False.

        To define a new type of geometry you need to use the
        `register_geometry` decorator. You are encouraged to do this for your
        own BOUT++ physics module, to apply relevant normalisations.
    gridfilepath : str, optional
        The path to a grid file, containing any variables needed to apply the geometry
        specified by the 'geometry' option, which are not contained in the dump files.
    keep_xboundaries : bool, optional
        If true, keep x-direction boundary cells (the cells past the physical
        edges of the grid, where boundary conditions are set); increases the
        size of the x dimension in the returned data-set. If false, trim these
        cells.
    keep_yboundaries : bool, optional
        If true, keep y-direction boundary cells (the cells past the physical
        edges of the grid, where boundary conditions are set); increases the
        size of the y dimension in the returned data-set. If false, trim these
        cells.
    run_name : str, optional
        Name to give to the whole dataset, e.g. 'JET_ELM_high_resolution'.
        Useful if you are going to open multiple simulations and compare the
        results.
    info : bool or "terse", optional
    is_restart : bool, optional
        Restart files require some special handling (e.g. working around variables that
        are not present in restart files). By default, this special handling is enabled
        if the files do not have a time dimension and `restart` is present in the file
        name in `datapath`. This option can be set to True or False to explicitly enable
        or disable the restart file handling.
    kwargs : optional
        Keyword arguments are passed down to `xarray.open_mfdataset`, which in
        turn passes extra kwargs down to `xarray.open_dataset`.

    Returns
    -------
    ds : xarray.Dataset
    """

    if chunks is None:
        chunks = {}

    input_type = _check_dataset_type(datapath)

    if is_restart is None:
        is_restart = input_type == "restart"
    elif is_restart is True:
        input_type = "restart"

    if "reload" in input_type:
        if input_type == "reload":
            if isinstance(datapath, Path):
                # xr.open_mfdataset only accepts glob patterns as strings, not Path
                # objects
                datapath = str(datapath)
            ds = xr.open_mfdataset(
                datapath,
                chunks=chunks,
                combine="by_coords",
                data_vars="minimal",
                **kwargs,
            )
        elif input_type == "reload_fake":
            ds = xr.combine_by_coords(datapath, data_vars="minimal").chunk(chunks)
        else:
            raise ValueError(f"internal error: unexpected input_type={input_type}")

        def attrs_to_dict(obj, section):
            result = {}
            section = section + ":"
            sectionlength = len(section)
            for key in list(obj.attrs):
                if key[:sectionlength] == section:
                    val = obj.attrs.pop(key)
                    if isinstance(val, bytes):
                        val = val.decode()
                    result[key[sectionlength:]] = val
            return result

        def attrs_remove_section(obj, section):
            section = section + ":"
            sectionlength = len(section)
            has_metadata = False
            for key in list(obj.attrs):
                if key[:sectionlength] == section:
                    has_metadata = True
                    del obj.attrs[key]
            return has_metadata

        # Restore metadata from attrs
        metadata = attrs_to_dict(ds, "metadata")
        if "is_restart" not in metadata:
            # Loading data that was saved with a version of xbout from before
            # "is_restart" was added, so need to add it to the metadata.
            metadata["is_restart"] = int(is_restart)
        ds.attrs["metadata"] = metadata
        # Must do this for all variables and coordinates in dataset too
        for da in chain(ds.data_vars.values(), ds.coords.values()):
            if attrs_remove_section(da, "metadata"):
                da.attrs["metadata"] = metadata

        ds = _add_options(ds, inputfilepath)

        # If geometry was set, apply geometry again
        if geometry is not None:
            if "geometry" != ds.attrs.get("geometry", None):
                warn(
                    f'open_boutdataset() called with geometry="{geometry}", but we are '
                    f"reloading a Dataset that was saved after being loaded with "
                    f'geometry="{ds.attrs.get("geometry", None)}". Applying '
                    f'geometry="{geometry}" from the argument.'
                )
            if gridfilepath is not None:
                grid = _open_grid(
                    gridfilepath,
                    chunks=chunks,
                    keep_xboundaries=ds.metadata["keep_xboundaries"],
                    keep_yboundaries=ds.metadata["keep_yboundaries"],
                    mxg=ds.metadata["MXG"],
                )
            else:
                grid = None
            ds = geometries.apply_geometry(ds, geometry, grid=grid)
        elif "geometry" in ds.attrs:
            ds = geometries.apply_geometry(ds, ds.attrs["geometry"])
        else:
            ds = geometries.apply_geometry(ds, None)

        if info == "terse":
            print("Read in dataset from {}".format(str(Path(datapath))))
        elif info:
            print("Read in:\n{}".format(ds.bout))

        return ds

    # Determine if file is a grid file or data dump files
    remove_yboundaries = False
    if "dump" in input_type or "restart" in input_type:
        # Gather pointers to all numerical data from BOUT++ output files
        ds, remove_yboundaries = _auto_open_mfboutdataset(
            datapath=datapath,
            chunks=chunks,
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
            is_restart=is_restart,
            **kwargs,
        )
    elif "grid" in input_type:
        # Its a grid file
        ds = _open_grid(
            datapath,
            chunks=chunks,
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
            **kwargs,
        )
    else:
        raise ValueError(f"internal error: unexpected input_type={input_type}")

    if not is_restart:
        for var in _BOUT_TIME_DEPENDENT_META_VARS:
            if var in ds:
                # Assume different processors in x & y have same iteration etc.
                latest_top_left = {dim: 0 for dim in ds[var].dims}
                if "t" in ds[var].dims:
                    latest_top_left["t"] = -1
                ds[var] = ds[var].isel(latest_top_left).squeeze(drop=True)

    ds, metadata = _separate_metadata(ds)
    # Store as ints because netCDF doesn't support bools, so we can't save
    # bool attributes
    metadata["keep_xboundaries"] = int(keep_xboundaries)
    metadata["keep_yboundaries"] = int(keep_yboundaries)
    metadata["is_restart"] = int(is_restart)
    ds = _set_attrs_on_all_vars(ds, "metadata", metadata)

    if remove_yboundaries:
        # If remove_yboundaries is True, we need to keep y-boundaries when opening the
        # grid file, as they will be removed from the full Dataset below
        keep_yboundaries = True

    ds = _add_options(ds, inputfilepath)

    if geometry is None:
        if geometry in ds.attrs:
            geometry = ds.attrs.get("geometry")
    if geometry:
        if info:
            print("Applying {} geometry conventions".format(geometry))

        if gridfilepath is not None:
            grid = _open_grid(
                gridfilepath,
                chunks=chunks,
                keep_xboundaries=keep_xboundaries,
                keep_yboundaries=keep_yboundaries,
                mxg=ds.metadata["MXG"],
            )
        else:
            grid = None
    else:
        grid = None
        if info:
            warn("No geometry type found, no physical coordinates will be added")

    # Update coordinates to match particular geometry of grid
    ds = geometries.apply_geometry(ds, geometry, grid=grid)

    if remove_yboundaries:
        ds = ds.bout.remove_yboundaries()

    # TODO read and store git commit hashes from output files

    if run_name:
        ds.name = run_name

    # Set some default settings that are only used in post-processing by xBOUT, not by
    # BOUT++
    ds.bout.fine_interpolation_factor = 8

    if info == "terse":
        print("Read in dataset from {}".format(str(Path(datapath))))
    elif info:
        print("Read in:\n{}".format(ds.bout))

    return ds


def _add_options(ds, inputfilepath):
    if inputfilepath:
        # Use Ben's options class to store all input file options
        options = BoutOptionsFile(
            inputfilepath,
            nx=ds.metadata["nx"],
            ny=ds.metadata["ny"],
            nz=ds.metadata["nz"],
        )
    else:
        options = None
    ds = _set_attrs_on_all_vars(ds, "options", options)
    return ds


def collect(
    varname,
    xind=None,
    yind=None,
    zind=None,
    tind=None,
    path=".",
    yguards=False,
    xguards=True,
    info=True,
    prefix="BOUT.dmp",
):
    """

    Extract the data pertaining to a specified variable in a BOUT++ data set


    Parameters
    ----------
    varname : str
        Name of the variable
    xind, yind, zind, tind : int, slice or list of int, optional
        Range of X, Y, Z or time indices to collect. Either a single
        index to collect, a list containing [start, end] (inclusive
        end), or a slice object (usual python indexing). Default is to
        fetch all indices
    path : str, optional
        Path to data files (default: ".")
    prefix : str, optional
        File prefix (default: "BOUT.dmp")
    yguards : bool, optional
        Collect Y boundary guard cells? (default: False)
    xguards : bool, optional
        Collect X boundary guard cells? (default: True)
        (Set to True to be consistent with the definition of nx)
    info : bool, optional
        Print information about collect? (default: True)

    Notes
    ----------
    strict : This option found in boutdata.collect() is not present in this function
             it is assumed that the varname given is correct, if variable does not exist
             the function will fail
    tind_auto : This option is not required when using _auto_open_mfboutdataset as an
             automatic failure if datasets are different lengths is included

    Returns
    ----------
    ds : numpy.ndarray

    """
    from os.path import join

    datapath = join(path, prefix + "*.nc")

    ds, _ = _auto_open_mfboutdataset(
        datapath, keep_xboundaries=xguards, keep_yboundaries=yguards, info=info
    )

    if varname not in ds:
        raise KeyError("No variable, {} was found in {}.".format(varname, datapath))

    dims = list(ds.dims)
    inds = [tind, xind, yind, zind]

    selection = {}

    # Convert indexing values to an isel suitable format
    for dim, ind in zip(dims, inds):

        if isinstance(ind, int):
            indexer = [ind]
        elif isinstance(ind, list):
            start, end = ind
            indexer = slice(start, end + 1)
        elif ind is not None:
            indexer = ind
        else:
            indexer = None

        if indexer:
            selection[dim] = indexer

    try:
        version = ds["BOUT_VERSION"]
    except KeyError:
        # If BOUT Version is not saved in the dataset
        version = 0

    # Subtraction of z-dimensional data occurs in boutdata.collect
    # if BOUT++ version is old - same feature added here
    if (version < 3.5) and ("z" in dims):
        zsize = int(ds["nz"]) - 1
        ds = ds.isel(z=slice(zsize))

    if selection:
        ds = ds.isel(selection)

    result = ds[varname].values

    # Close netCDF files to ensure they are not locked if collect is called again
    ds.close()

    return result


def _check_dataset_type(datapath):
    """
    Check what type of files we have. Could be:
    (i) produced by xBOUT
        - one or several files, include metadata attributes, e.g.
          'metadata:keep_yboundaries'
    (ii) grid file
        - only one file, and no time dimension
    (iii) produced by BOUT++
        - one or several files
    (iv) restart files produced by BOUT++
        - one or several files, no time dimension, filenames include `restart`
    """

    if not _is_path(datapath):
        # not a filepath glob, so presumably Dataset or list of Datasets used for
        # testing
        if isinstance(datapath, xr.Dataset):
            if "metadata:keep_yboundaries" in datapath.attrs:
                # (i)
                return "reload_fake"
            elif "t" in datapath.dims:
                # (iii)
                return "dump_fake"
            else:
                # (ii)
                return "grid_fake"
        elif len(datapath) > 1:
            if "metadata:keep_yboundaries" in datapath[0].attrs:
                # (i)
                return "reload_fake"
            else:
                # (iii)
                return "dump_fake"
        else:
            # Single element list of Datasets, or nested list of Datasets
            return _check_dataset_type(datapath[0])

    filepaths, filetype = _expand_filepaths(datapath)

    ds = xr.open_dataset(filepaths[0], engine=filetype)
    ds.close()
    if "metadata:keep_yboundaries" in ds.attrs:
        # (i)
        return "reload"
    elif "t" in ds.dims:
        # (iii)
        return "dump"
    elif all(["restart" in Path(p).name for p in filepaths]):
        # (iv)
        return "restart"
    elif len(filepaths) == 1:
        # (ii)
        return "grid"
    else:
        # fall back to opening as dump files
        return "dump"


def _auto_open_mfboutdataset(
    datapath,
    chunks=None,
    info=True,
    keep_xboundaries=False,
    keep_yboundaries=False,
    is_restart=False,
    **kwargs,
):
    if chunks is None:
        chunks = {}

    if is_restart:
        data_vars = "minimal"
    else:
        data_vars = _BOUT_TIME_DEPENDENT_META_VARS

    if _is_path(datapath):
        filepaths, filetype = _expand_filepaths(datapath)

        # Open just one file to read processor splitting
        nxpe, nype, mxg, myg, mxsub, mysub, is_squashed_doublenull = _read_splitting(
            filepaths[0], info, keep_yboundaries
        )

        if is_squashed_doublenull:
            # Need to remove y-boundaries after loading: (i) in case we are loading a
            # squashed data-set, in which case we cannot easily remove the upper
            # boundary cells in _trim(); (ii) because using the remove_yboundaries()
            # method for non-squashed data-sets is simpler than replicating that logic
            # in _trim().
            remove_yboundaries = not keep_yboundaries
            keep_yboundaries = True
        else:
            remove_yboundaries = False

        _preprocess = partial(
            _trim,
            guards={"x": mxg, "y": myg},
            keep_boundaries={"x": keep_xboundaries, "y": keep_yboundaries},
            nxpe=nxpe,
            nype=nype,
            is_restart=is_restart,
        )

        paths_grid, concat_dims = _arrange_for_concatenation(filepaths, nxpe, nype)

        ds = xr.open_mfdataset(
            paths_grid,
            concat_dim=concat_dims,
            combine="nested",
            data_vars=data_vars,
            preprocess=_preprocess,
            engine=filetype,
            chunks=chunks,
            join="exact",
            **kwargs,
        )
    else:
        # datapath was nested list of Datasets

        if isinstance(datapath, xr.Dataset):
            # normalise as one-element list
            datapath = [datapath]

        mxg = int(datapath[0]["MXG"])
        myg = int(datapath[0]["MYG"])
        nxpe = int(datapath[0]["NXPE"])
        nype = int(datapath[0]["NYPE"])
        is_squashed_doublenull = (
            len(datapath) == 1
            and (datapath[0]["jyseps2_1"] != datapath[0]["jyseps1_2"]).values
        )

        if is_squashed_doublenull:
            # Need to remove y-boundaries after loading when loading a squashed
            # data-set, in which case we cannot easily remove the upper boundary cells
            # in _trim().
            remove_yboundaries = not keep_yboundaries
            keep_yboundaries = True
        else:
            remove_yboundaries = False

        _preprocess = partial(
            _trim,
            guards={"x": mxg, "y": myg},
            keep_boundaries={"x": keep_xboundaries, "y": keep_yboundaries},
            nxpe=nxpe,
            nype=nype,
            is_restart=is_restart,
        )

        datapath = [_preprocess(x) for x in datapath]

        ds_grid, concat_dims = _arrange_for_concatenation(datapath, nxpe, nype)

        ds = xr.combine_nested(
            ds_grid,
            concat_dim=concat_dims,
            data_vars=data_vars,
            join="exact",
            combine_attrs="no_conflicts",
        )

    if not is_restart:
        # Remove any duplicate time values from concatenation
        _, unique_indices = unique(ds["t_array"], return_index=True)
        ds = ds.isel(t=unique_indices)

    return ds, remove_yboundaries


def _expand_filepaths(datapath):
    """Determines filetypes and opens all dump files."""
    path = Path(datapath)

    filetype = _check_filetype(path)

    filepaths = _expand_wildcards(path)

    if not filepaths:
        raise IOError("No datafiles found matching datapath={}".format(datapath))

    if len(filepaths) > 128:
        warn(
            "Trying to open a large number of files - setting xarray's"
            " `file_cache_maxsize` global option to {} to accommodate this. "
            "Recommend using `xr.set_options(file_cache_maxsize=NUM)`"
            " to explicitly set this to a large enough value.".format(
                str(len(filepaths))
            )
        )
        xr.set_options(file_cache_maxsize=len(filepaths))

    return filepaths, filetype


def _expand_wildcards(path):
    """Return list of filepaths matching wildcard"""

    # Find first parent directory which does not contain a wildcard
    base_dir = Path(path.anchor)

    # Find path relative to parent
    search_pattern = str(path.relative_to(base_dir))

    # Search this relative path from the parent directory for all files matching user input
    filepaths = list(base_dir.glob(search_pattern))

    # Sort by numbers in filepath before returning
    # Use "natural" sort to avoid lexicographic ordering of numbers
    # e.g. ['0', '1', '10', '11', etc.]
    return natsorted(filepaths, key=lambda filepath: str(filepath))


def _read_splitting(filepath, info, keep_yboundaries):
    ds = xr.open_dataset(str(filepath))

    # Account for case of no parallelisation, when nxpe etc won't be in dataset
    def get_nonnegative_scalar(ds, key, default=1, info=True):
        if key in ds:
            val = ds[key].values
            if val < 0:
                raise ValueError(
                    f"{key} read from dump files is {val}, but negative"
                    f" values are not valid"
                )
            else:
                return val
        else:
            if info is True:
                print(f"{key} not found, setting to {default}")
            if default < 0:
                raise ValueError(
                    f"Default for {key} is {val}, but negative values are not valid"
                )
            return default

    nxpe = get_nonnegative_scalar(ds, "NXPE", default=1, info=info)
    nype = get_nonnegative_scalar(ds, "NYPE", default=1, info=info)
    mxg = get_nonnegative_scalar(ds, "MXG", default=2, info=info)
    myg = get_nonnegative_scalar(ds, "MYG", default=0, info=info)
    mxsub = get_nonnegative_scalar(
        ds, "MXSUB", default=ds.dims["x"] - 2 * mxg, info=info
    )
    mysub = get_nonnegative_scalar(
        ds, "MYSUB", default=ds.dims["y"] - 2 * myg, info=info
    )

    # Check whether this is a single file squashed from the multiple output files of a
    # parallel run (i.e. NXPE*NYPE > 1 even though there is only a single file to read).
    nx = ds["nx"].values
    ny = ds["ny"].values
    nx_file = ds.dims["x"]
    ny_file = ds.dims["y"]
    is_squashed_doublenull = False
    if nxpe > 1 or nype > 1:
        # if nxpe = nype = 1, was only one process anyway, so no need to check for
        # squashing
        if nx_file == nx or nx_file == nx - 2 * mxg:
            has_xboundaries = nx_file == nx
            if not has_xboundaries:
                mxg = 0

            # Check if there are two divertor targets present
            if ds["jyseps1_2"] > ds["jyseps2_1"]:
                upper_target_cells = myg
            else:
                upper_target_cells = 0
            if ny_file == ny or ny_file == ny + 2 * myg + 2 * upper_target_cells:
                # This file contains all the points, possibly including guard cells

                has_yboundaries = not (ny_file == ny)
                if not has_yboundaries:
                    myg = 0

                nxpe = 1
                nype = 1
                is_squashed_doublenull = (ds["jyseps2_1"] != ds["jyseps1_2"]).values
            elif ny_file == ny + 2 * myg:
                # Older squashed file from double-null grid but containing only lower
                # target boundary cells.
                if keep_yboundaries:
                    raise ValueError(
                        "Cannot keep y-boundary points: squashed file is missing upper "
                        "target boundary points."
                    )
                has_yboundaries = not (ny_file == ny)
                if not has_yboundaries:
                    myg = 0

                nxpe = 1
                nype = 1
                # For this case, do not need the special handling enabled by
                # is_squashed_doublenull=True, as keeping y-boundaries is not allowed
                is_squashed_doublenull = False

    # Avoid trying to open this file twice
    ds.close()

    return nxpe, nype, mxg, myg, mxsub, mysub, is_squashed_doublenull


def _arrange_for_concatenation(filepaths, nxpe=1, nype=1):
    """
    Arrange filepaths into a nested list-of-lists which represents their
    ordering across different processors and consecutive simulation runs.

    Filepaths must be a sorted list. Uses the fact that BOUT's output files are
    named as num = nxpe*i + j, where i={0, ..., nype}, j={0, ..., nxpe}.
    Also assumes that any consecutive simulation runs are in directories which
    when sorted are in the correct order (e.g. /run0/*, /run1/*, ...).
    """

    nprocs = nxpe * nype
    n_runs = int(len(filepaths) / nprocs)
    if len(filepaths) % nprocs != 0:
        raise ValueError(
            "Each run directory does not contain an equal number "
            "of output files. If the parallelization scheme of "
            "your simulation changed partway-through, then please "
            "load each directory separately and concatenate them "
            "along the time dimension with xarray.concat()."
        )

    # Create list of lists of filepaths, so that xarray knows how they should
    # be concatenated by xarray.open_mfdataset()
    paths = iter(filepaths)
    paths_grid = [
        [[next(paths) for x in range(nxpe)] for y in range(nype)] for t in range(n_runs)
    ]

    # Dimensions along which no concatenation is needed are still present as
    # single-element lists, so need to concatenation along dim=None for those
    concat_dims = [None, None, None]
    if len(filepaths) > nprocs:
        concat_dims[0] = "t"
    if nype > 1:
        concat_dims[1] = "y"
    if nxpe > 1:
        concat_dims[2] = "x"

    return paths_grid, concat_dims


def _trim(ds, *, guards, keep_boundaries, nxpe, nype, is_restart):
    """
    Trims all guard (and optionally boundary) cells off a single dataset read from a
    single BOUT dump file, to prepare for concatenation.
    Also drops some variables that store timing information, which are different for each
    process and so cannot be concatenated.

    Parameters
    ----------
    guards : dict
        Number of guard cells along each dimension, e.g. {'x': 2, 't': 0}
    keep_boundaries : dict
        Whether or not to preserve the boundary cells along each dimension, e.g.
        {'x': True, 'y': False}
    nxpe : int
        Number of processors in x direction
    nype : int
        Number of processors in y direction
    is_restart : bool
        Is data being loaded from restart files?
    """

    if any(keep_boundaries.values()):
        # Work out if this particular dataset contains any boundary cells
        lower_boundaries, upper_boundaries = _infer_contains_boundaries(ds, nxpe, nype)
    else:
        lower_boundaries, upper_boundaries = {}, {}

    selection = {}
    for dim in ds.dims:
        lower = _get_limit("lower", dim, keep_boundaries, lower_boundaries, guards)
        upper = _get_limit("upper", dim, keep_boundaries, upper_boundaries, guards)
        selection[dim] = slice(lower, upper)
    trimmed_ds = ds.isel(**selection)

    # Ignore FieldPerps for now
    for name in trimmed_ds:
        if trimmed_ds[name].dims == ("x", "z") or trimmed_ds[name].dims == (
            "t",
            "x",
            "z",
        ):
            trimmed_ds = trimmed_ds.drop_vars(name)

    to_drop = _BOUT_PER_PROC_VARIABLES

    return trimmed_ds.drop_vars(to_drop, errors="ignore")


def _infer_contains_boundaries(ds, nxpe, nype):
    """
    Uses the processor indices and BOUT++'s topology indices to work out whether this
    dataset contains boundary cells, and on which side.
    """

    if nxpe * nype == 1:
        # single file, always contains boundaries
        return {"x": True, "y": True}, {"x": True, "y": True}

    try:
        xproc = int(ds["PE_XIND"])
        yproc = int(ds["PE_YIND"])
    except KeyError:
        # output file from BOUT++ earlier than 4.3
        # Use knowledge that BOUT names its output files as /folder/prefix.num.nc, with a
        # numbering scheme
        # num = nxpe*i + j, where i={0, ..., nype}, j={0, ..., nxpe}
        filename = ds.encoding["source"]
        *prefix, filenum, extension = Path(filename).suffixes
        filenum = int(filenum.replace(".", ""))
        xproc = filenum % nxpe
        yproc = filenum // nxpe

    lower_boundaries, upper_boundaries = {}, {}

    lower_boundaries["x"] = xproc == 0
    upper_boundaries["x"] = xproc == nxpe - 1

    lower_boundaries["y"] = yproc == 0
    upper_boundaries["y"] = yproc == nype - 1

    jyseps2_1 = int(ds["jyseps2_1"])
    jyseps1_2 = int(ds["jyseps1_2"])
    if jyseps1_2 > jyseps2_1:
        # second divertor present
        ny_inner = int(ds["ny_inner"])
        mysub = int(ds["MYSUB"])
        if mysub * (yproc + 1) == ny_inner:
            upper_boundaries["y"] = True
        elif mysub * yproc == ny_inner:
            lower_boundaries["y"] = True

    return lower_boundaries, upper_boundaries


def _get_limit(side, dim, keep_boundaries, boundaries, guards):
    # Check for boundary cells, otherwise use guard cells, else leave alone

    if keep_boundaries.get(dim, False):
        if boundaries.get(dim, False):
            limit = None
        else:
            limit = guards[dim] if side == "lower" else -guards[dim]
    elif guards.get(dim, False):
        limit = guards[dim] if side == "lower" else -guards[dim]
    else:
        limit = None

    if limit == 0:
        # 0 would give incorrect result as an upper limit
        limit = None
    return limit


def _open_grid(datapath, chunks, keep_xboundaries, keep_yboundaries, mxg=2, **kwargs):
    """
    Opens a single grid file. Implements slightly different logic for
    boundaries to deal with different conventions in a BOUT grid file.
    """

    acceptable_dims = ["x", "y", "z"]

    # Passing 'chunks' with dimensions that are not present in the dataset causes an
    # error. A gridfile will be missing 't' and may be missing 'z' dimensions that dump
    # files have, so we must remove them from 'chunks'.
    grid_chunks = copy(chunks)
    unrecognised_chunk_dims = list(set(grid_chunks.keys()) - set(acceptable_dims))
    for dim in unrecognised_chunk_dims:
        del grid_chunks[dim]

    if _is_path(datapath):
        gridfilepath = Path(datapath)
        grid = xr.open_dataset(
            gridfilepath, engine=_check_filetype(gridfilepath), **kwargs
        )
    else:
        grid = datapath

    # TODO find out what 'yup_xsplit' etc are in the doublenull storm file John gave me
    # For now drop any variables with extra dimensions
    unrecognised_dims = list(set(grid.dims) - set(acceptable_dims))
    if len(unrecognised_dims) > 0:
        # Weird string formatting is a workaround to deal with possible bug in
        # pytest warnings capture - doesn't match strings containing brackets
        warn(
            "Will drop all variables containing the dimensions {} because "
            "they are not recognised".format(str(unrecognised_dims)[1:-1])
        )
        grid = grid.drop_dims(unrecognised_dims)

    if keep_xboundaries:
        # Set MXG so that it is picked up in metadata - needed for applying geometry,
        # etc.
        grid["MXG"] = mxg
    else:
        xboundaries = mxg
        if xboundaries > 0:
            grid = grid.isel(x=slice(xboundaries, -xboundaries, None))
        # Set MXG so that it is picked up in metadata - needed for applying geometry,
        # etc.
        grid["MXG"] = 0
    try:
        yboundaries = int(grid["y_boundary_guards"])
    except KeyError:
        # y_boundary_guards variable not in grid file - older grid files
        # never had y-boundary cells
        yboundaries = 0
    if keep_yboundaries:
        # Set MYG so that it is picked up in metadata - needed for applying geometry,
        # etc.
        grid["MYG"] = yboundaries
    else:
        if yboundaries > 0:
            # Remove y-boundary cells from first divertor target
            grid = grid.isel(y=slice(yboundaries, -yboundaries, None))
            if grid["jyseps1_2"] > grid["jyseps2_1"]:
                # There is a second divertor target, remove y-boundary cells
                # there too
                nin = int(grid["ny_inner"])
                grid_lower = grid.isel(y=slice(None, nin, None))
                grid_upper = grid.isel(y=slice(nin + 2 * yboundaries, None, None))
                grid = xr.concat(
                    (grid_lower, grid_upper),
                    dim="y",
                    data_vars="minimal",
                    compat="identical",
                    join="exact",
                )
        # Set MYG so that it is picked up in metadata - needed for applying geometry,
        # etc.
        grid["MYG"] = 0

    if "z" in grid_chunks and "z" not in grid.dims:
        del grid_chunks["z"]
    grid = grid.chunk(grid_chunks)

    return grid
